from django.db.models import IntegerField
from django.db.models.lookups import Transform


class HLLHash(Transform):
    """
    Default hash function for HLL
    - it hashes and transforms any field to an unsigned int 32
    """

    function = "hll_hash"
    lookup_name = "hll_hash"
    output_field = IntegerField()



from django.db import migrations

from . import load_sql


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.RunSQL(
            sql=load_sql(__file__), reverse_sql=load_sql(__file__, reverse=True)
        )
    ]



from django.db.models import Aggregate, IntegerField


class HLLCardinality(Aggregate):
    """
    Return an approximate distinct count based on the HyperLogLog algorithm
    as described in:
        http://algo.inria.fr/flajolet/Publications/FlMa85.pdf

    Despite being written in a combination of PL/pgSQL and SQL, it should
    outperform COUNT(DISTINCT ...) in many circumstances. It should also be
    more memory efficient. Notice, however that it will always be an approximate
    result rather than an exact count.
    """

    function = "hll_cardinality"
    name = "HLLCardinality"
    allow_distinct = False
    output_field = IntegerField()
    empty_result_set_value = 0


class HLLCardinalityFromHash(Aggregate):
    """
    Return an approximate distinct count based on the HyperLogLog algorithm
    as described in:
        http://algo.inria.fr/flajolet/Publications/FlMa85.pdf

    Requires the input to be previously hashed, see `functions.HLLHash`

    Note that the chosen hash function needs to produce a 32 bit uniformly
    distributed unsigned integers for the approximation to be accurate.

    MD5 and SHA1 both produce uniform hashes, but they would have to be cast into
    an integer to be used with this function. This is not recommended as it would
    be very slow.
    """

    function = "hll_cardinality_from_hash"
    name = "HLLCardinalityFromHash"
    allow_distinct = False
    output_field = IntegerField()
    empty_result_set_value = 0



from os import PathLike
from pathlib import Path
from typing import Union


def load_sql(path: Union[PathLike[str], str], reverse: bool = False) -> str:
    """
    This utility loads sql from base path with a `.sql` suffix,
    optionally it attaches a `.reverse.sql` suffix

    We use it to write migrations that can be run both forwards and backwards

    e.g.
    load_sql('migrations/0001_initial'),
        will load `migrations/0001_initial.sql`

    and

    load_sql('migrations/0001_initial', reverse=True),
        will load `migrations/0001_initial.reverse.sql`

    """
    path = Path(path)
    parent = path.parent
    stem = path.stem
    if reverse:
        filename = f"{stem}.reverse.sql"
    else:
        filename = f"{stem}.sql"

    with open(parent / filename) as sql_file:
        return sql_file.read()
